// isosurface table

// Copyright (c) 2001 by Rephael Wenger, The Ohio State University 

#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "isotable.h"

using namespace ISOTABLE;

#ifndef LONG_BIT

#define LONG_BIT (CHAR_BIT * sizeof(long))

#endif

//**************************************************
// ISOSURFACE_TABLE_POLYHEDRON
//**************************************************

ISOSURFACE_TABLE_POLYHEDRON::ISOSURFACE_TABLE_POLYHEDRON(const int d)
// constructor
{
  Init();
  dimension = d;
}

ISOSURFACE_TABLE_POLYHEDRON::~ISOSURFACE_TABLE_POLYHEDRON()
// destructor
{
  FreeAll();
}

ISOSURFACE_TABLE_POLYHEDRON::ISOSURFACE_TABLE_POLYHEDRON
  (const ISOSURFACE_TABLE_POLYHEDRON & init)
// copy
{
  Init();

  *this = init;
}

const ISOSURFACE_TABLE_POLYHEDRON & ISOSURFACE_TABLE_POLYHEDRON::operator = 
  (const ISOSURFACE_TABLE_POLYHEDRON & right)  
// assign 
{
  if (&right != this) {         // avoid self-assignment
    FreeAll();
    dimension = right.Dimension();
    SetSize(right.NumVertices(), right.NumEdges());

    // copy vertices
    for (int iv = 0; iv < NumVertices(); iv++)
      for (int ic = 0; ic < Dimension(); ic++)
	SetVertexCoord(iv, ic, right.VertexCoord(iv, ic));

    // copy edges
    for (int ie = 0; ie < NumVertices(); ie++)
      SetEdge(ie, right.EdgeEndpoint(ie, 0), right.EdgeEndpoint(ie, 1));
  };

  return *this;
}

void ISOSURFACE_TABLE_POLYHEDRON::FreeAll()
// free all memory
{
  num_vertices = 0;
  num_edges = 0;
  delete [] vertex_coord;
  vertex_coord = NULL;
  delete [] edge_endpoint;
  edge_endpoint = NULL;
}

void ISOSURFACE_TABLE_POLYHEDRON::Init()
// initialize
{
  dimension = 0;
  num_vertices = 0;
  num_edges = 0;
  vertex_coord = NULL;
  edge_endpoint = NULL;
}

bool ISOSURFACE_TABLE_POLYHEDRON::VertexSign
  (const long code, const int iv) const
// sign of vertex iv in code
{
  long mask = 1L << iv;

  if ((mask & code) == 0)
    return(false);
  else
    return(true);
}

void ISOSURFACE_TABLE_POLYHEDRON::SetDimension(const int d)
// set polyhedron dimension
{
  FreeAll();
  num_vertices = num_edges = 0;

  dimension = d;
}

void ISOSURFACE_TABLE_POLYHEDRON::SetNumVertices(const int numv)
// set number of vertices
// Must be called before setting polyhedron vertices
{
  if (!CheckDimension())
    throw ISOSURFACE_TABLE_ERROR("Illegal polyhedron dimension.");

  FreeAll();
  num_vertices = num_edges = 0;

  if (numv == 0)
    throw ISOSURFACE_TABLE_ERROR("Number of vertices must be non-zero.");

  // Note that even if numv <= LONG_BIT, there may not be enough 
  //   memory to store the isosurface table.
  if (numv > LONG_BIT)
    throw ISOSURFACE_TABLE_ERROR
      ("Number of polyhedron vertices is too large.");

  num_vertices = numv;
  vertex_coord = new int[num_vertices*Dimension()];
}

void ISOSURFACE_TABLE_POLYHEDRON::SetNumEdges(const int nume)
// set number of edges
// Must be called before setting polyhedron edges
{
  delete [] edge_endpoint;
  edge_endpoint = NULL;
  num_edges = 0;

  if (!CheckDimension())
    throw ISOSURFACE_TABLE_ERROR("Illegal dimension.");

  if (NumVertices() == 0)
    throw ISOSURFACE_TABLE_ERROR
      ("Number of vertices must be set before number of edges.");

  if (nume < 1)
    throw ISOSURFACE_TABLE_ERROR("Number of edges must be non-zero.");

  num_edges = nume;
  edge_endpoint = new int[num_edges*2];
}


void ISOSURFACE_TABLE_POLYHEDRON::SetVertexCoord
  (const int iv, const int ic, const int coord)
// set polyhedron vertex coordinate
// iv = vertex index.  In range [0..NumVertices()-1].
// ic = coordinate index. In range [0..Dimension()-1].
// coord = coordinate.  Must be even.
{
  if (iv < 0 || iv >= NumVertices())
    throw ISOSURFACE_TABLE_ERROR("Illegal vertex index.");

  if (ic < 0 || ic >= Dimension())
    throw ISOSURFACE_TABLE_ERROR("Illegal vertex coordinate index.");

  if (coord%2 != 0)
    throw ISOSURFACE_TABLE_ERROR
      ("Illegal vertex coordinate.  Vertex coordinate must be even.");

  if (vertex_coord == NULL)
    throw ISOSURFACE_TABLE_ERROR
      ("Vertex coordinate memory not allocated.");

  vertex_coord[iv*Dimension() + ic] = coord;
}

void ISOSURFACE_TABLE_POLYHEDRON::SetEdge
  (const int ie, const int iv0, const int iv1)
// set polyhedron edge coordinate
// ie = edge index.  In range [0..NumEdges()-1].
// iv0 = vertex 0 index.  In range [0..NumVertices()-1].
// iv1 = vertex 1 index.  In range [0..NumVertices()-1].
{
  if (ie < 0 || ie >= NumEdges())
    throw ISOSURFACE_TABLE_ERROR("Illegal edge index.");

  if (iv0 < 0 || iv0 > NumVertices() ||
      iv1 < 0 || iv1 > NumVertices())
    throw ISOSURFACE_TABLE_ERROR("Illegal vertex index.");

  if (edge_endpoint == NULL)
    throw ISOSURFACE_TABLE_ERROR("Edge endpoint memory not allocated.");

  edge_endpoint[2*ie] = iv0;
  edge_endpoint[2*ie+1] = iv1;
}

int ISOSURFACE_TABLE_POLYHEDRON::MidpointCoord
  (const int ie, const int ic) const
// return ic'th coordinate of midpoint of edge ie
// Note: vertice coordinates must all be even so midpoint coordinate
//   is an integer
{
  int iv0 = EdgeEndpoint(ie, 0);
  int iv1 = EdgeEndpoint(ie, 1);
  int coord0 = VertexCoord(iv0, ic);
  int coord1 = VertexCoord(iv1, ic);
  
  return((coord0+coord1)/2);
}

void ISOSURFACE_TABLE_POLYHEDRON::GenCube(const int cube_dimension)
// generate a polyhedron
{
  dimension = cube_dimension;

  if (!CheckDimension())
    throw ISOSURFACE_TABLE_ERROR();

  int numv = 1L << Dimension();
  int nume = (numv * Dimension()) / 2;

  SetSize(numv, nume);

  // set vertex coordinates
  for (int iv = 0; iv < NumVertices(); iv++) {
    long mask = 1L;
    for (int ic = 0; ic < Dimension(); ic++) {
      int bit = mask & iv;
      int coord = 0;
      if (bit != 0)
	coord = 2;
      SetVertexCoord(iv, ic, coord);
      mask = mask << 1;
    };
  };

  // generate edges in lexicographic order
  int ie = 0;
  long control = 0;
  while (ie < NumEdges()) {

    // find first 0 bit in control
    int ic = 0;
    long mask = 1L;
    while ((mask & control) != 0) {
      mask = mask << 1;
      ic++;
    };

    // find start vertex by stripping ic bits from control
    int start = control;
    start = start >> ic;
    start = start << ic;

    for (int i = 0; i < (1L << ic); i++) {
      int iv0 = start + i;
      int iv1 = iv0 + (1L << ic);
      SetEdge(ie, iv0, iv1);
      ie++;
    };

    control++;
  };

  if (control+1 != (1L << Dimension()))
    throw ISOSURFACE_TABLE_ERROR("Programming error detected in GenCube.");
}


void ISOSURFACE_TABLE_POLYHEDRON::GenSimplex(const int simplex_dimension)
// generate a simplex
{
  dimension = simplex_dimension;

  if (!CheckDimension())
    throw ISOSURFACE_TABLE_ERROR();

  int numv = Dimension() + 1;
  int nume = (numv * Dimension()) / 2;

  SetSize(numv, nume);

  // initialize all vertex coordinates to 0
  for (int iv = 0; iv < NumVertices(); iv++)
    for (int ic = 0; ic < Dimension(); ic++)
      SetVertexCoord(iv, ic, 0);

  // set vertex coordinates
  int ic = 0;
  for (int jv = 1; jv < NumVertices(); jv++) {
    SetVertexCoord(jv, ic, 2);
    ic++;
  };

  // generate edges in lexicographic order
  int ie = 0;
  for (int iv0 = 0; iv0 < NumVertices()-1; iv0++)
    for (int iv1 = iv0+1; iv1 < NumVertices(); iv1++) {
      SetEdge(ie, iv0, iv1);
      ie++;
    };

  if (ie != nume)
    throw ISOSURFACE_TABLE_ERROR("Programming error detected in GenSimplex.");
}


void ISOSURFACE_TABLE_POLYHEDRON::WriteVertexCoordText(ostream & out) const
// write vertex coordinates in text format
{
  for (int iv = 0; iv < NumVertices(); iv++) {
    for (int ic = 0; ic < Dimension(); ic++) {
      out << "  " << VertexCoord(iv, ic);
    };
    out << endl;
  };
}

void ISOSURFACE_TABLE_POLYHEDRON::WriteEdgeListText(ostream & out) const
// write edge list in text format
{
  for (int ie = 0; ie < NumEdges(); ie++) {
    int iv0 = EdgeEndpoint(ie, 0);
    int iv1 = EdgeEndpoint(ie, 1);
    out << "  " << iv0 << "  " << iv1 << endl;
  };
}

void ISOSURFACE_TABLE_POLYHEDRON::ReadVertexCoordText(istream & in)
// read vertex coordinates in text format
{
  for (int iv = 0; iv < NumVertices(); iv++) {
    for (int ic = 0; ic < Dimension(); ic++) {
      if (in.eof())
	throw ISOSURFACE_TABLE_ERROR
	  ("Premature end of file while reading polyhedron vertices.");
      if (!in)
	throw ISOSURFACE_TABLE_ERROR("Error reading polyhedron vertices.");

      int coord;
      in >> coord;

      if (coord%2 != 0)
	throw ISOSURFACE_TABLE_ERROR
	  ("Polyhedron coordinates must be even.");

      SetVertexCoord(iv, ic, coord);
    };
  };
}

void ISOSURFACE_TABLE_POLYHEDRON::ReadEdgeListText(istream & in)
// read edge endpoints in text format
{
  for (int ie = 0; ie < NumEdges(); ie++) {
    if (in.eof())
      throw ISOSURFACE_TABLE_ERROR
	("Premature end of file while reading polyhedron edges.");
    if (!in)
      throw ISOSURFACE_TABLE_ERROR("Error reading polyhedron edges.");

    int iv0, iv1;
    in >> iv0 >> iv1;

    if (iv0 < 0 || iv1 < 0 || iv0 >= NumVertices() || iv1 >= NumVertices())
      throw ISOSURFACE_TABLE_ERROR
	("Illegal vertex index in polyhedron edge list.");

    SetEdge(ie, iv0, iv1);
  };
}

bool ISOSURFACE_TABLE_POLYHEDRON::CheckDimension() const
// check dimension
{
  if (dimension < 1)
    return(false);
  else
    return(true);
}

bool ISOSURFACE_TABLE_POLYHEDRON::Check() const
// check polyhedron
{
  if (!CheckDimension())
    return(false);

  if (NumVertices() < 1 || NumEdges() < 1 ||
      vertex_coord == NULL ||
      edge_endpoint == NULL)
    return(false);

  for (int iv = 0; iv < NumVertices(); iv++) {
    for (int ic = 0; ic < Dimension(); ic++) {
      if ((VertexCoord(iv, ic) % 2) != 0)
	return(false);
    };
  };

  for (int ie = 0; ie < NumEdges(); ie++) {
    for (int ip = 0; ip < 2; ip++) {
      int iv = EdgeEndpoint(ie, ip);
      if (iv < 0 || iv >= NumVertices())
	return(false);
    };
  };

  return(true);
}

//**************************************************
// ISOSURFACE_TABLE
//**************************************************

ISOSURFACE_TABLE::ISOSURFACE_TABLE_ENTRY::ISOSURFACE_TABLE_ENTRY()
// constructor
{
  num_simplices = 0;
  simplex_vertex_list = NULL;
}

ISOSURFACE_TABLE::ISOSURFACE_TABLE_ENTRY::~ISOSURFACE_TABLE_ENTRY()
// destructor
{
  delete [] simplex_vertex_list;
  simplex_vertex_list = NULL;
  num_simplices = 0;
}

bool ISOSURFACE_TABLE::ISOSURFACE_TABLE_ENTRY::Check() const
{
  if (num_simplices < 0)
    return(false);

  if (num_simplices > 0 && simplex_vertex_list == NULL)
    return(false);

  return(true);
}

// initializer for isosurface table file header
const char * const ISOSURFACE_TABLE::FILE_HEADER = "<ISOTABLE>";

ISOSURFACE_TABLE::ISOSURFACE_TABLE(const int d) :
  polyhedron(d)
// constructor
// d = dimension of space containing isosurface.  Should be 2, 3 or 4.
{
  max_num_vertices = 20;
  // Note: Even tables for polyhedron of this size are probably impossible 
  //   to compute/store

  num_table_entries = 0;
  entry = NULL;
  is_table_allocated = false;
  if (!CheckDimension())
    throw ISOSURFACE_TABLE_ERROR();
}

ISOSURFACE_TABLE::~ISOSURFACE_TABLE()
// destructor
{
  num_table_entries = 0;
  delete [] entry;
  entry = NULL;
}

void ISOSURFACE_TABLE::AllocTable()
// allocate memory for isosurface table
{
  if (Polyhedron().Dimension() == 0 ||
      Polyhedron().NumVertices() == 0 ||
      Polyhedron(). NumEdges() == 0) {
    throw ISOSURFACE_TABLE_ERROR
      ("Create polyhedron before allocating isosurface table.");
  };
      
  if (IsTableAllocated() || entry != NULL) {
    throw ISOSURFACE_TABLE_ERROR
      ("Isosurface table cannot be allocated more than once.");
  };

  delete [] entry;
  entry = NULL;
  num_table_entries = 0;

  if (Polyhedron().NumVertices() > MaxNumVertices())
    throw ISOSURFACE_TABLE_ERROR
      ("Table polyhedron contains too many vertices.");

  // use shift operator to calculate 2^(Polyhedron().NumVertices())
  num_table_entries = 1L << Polyhedron().NumVertices();
  entry = new ISOSURFACE_TABLE_ENTRY[num_table_entries];

  is_table_allocated = true;
}


void ISOSURFACE_TABLE::CheckTableAlloc() const
// throw error if table is allocated
{
  if (IsTableAllocated()) {
    throw ISOSURFACE_TABLE_ERROR
      ("Illegal attempt to define polyhedron after allocating isosurface table.");
  };
}

void ISOSURFACE_TABLE::SetNumSimplices(const int it, const int nums)
// set number of simplices in table entry it
// it = table entry
// nums = number of simplices
{
  if (!IsTableAllocated() || entry == NULL) {
    throw ISOSURFACE_TABLE_ERROR
      ("Table must be allocated before entering table entries.");
  };

  if (it < 0 || it >= NumTableEntries() ||
      nums < 0)
    throw ISOSURFACE_TABLE_ERROR();

  entry[it].num_simplices = 0;
  delete entry[it].simplex_vertex_list;
  entry[it].simplex_vertex_list = NULL;

  if (nums > 0)
    entry[it].simplex_vertex_list = new int[nums*NumVerticesPerSimplex()];

  entry[it].num_simplices = nums;
}


void ISOSURFACE_TABLE::SetSimplexVertex
  (const int it, const int is, const int iv, const int ie)
// set simplex vertex
// it = index table entry.  In range [0..NumTableEntries()-1].
// is = index simplex.  
// iv = index simplex vertex.  In range [0..NumVerticesPerSimplex()-1].
// ie = index of edge containing simplex vertex
{
  entry[it].simplex_vertex_list[is*NumVerticesPerSimplex()+iv] = ie;
}

void ISOSURFACE_TABLE::WriteText(ostream & out)
// write polyhedron & table, text format
{
  WriteFileHeaderText(out);
  out << Dimension() << endl;
  out << polyhedron.NumVertices() << endl;
  polyhedron.WriteVertexCoordText(out);
  out << polyhedron.NumEdges() << endl;
  polyhedron.WriteEdgeListText(out);
  WriteTableText(out);
}

void ISOSURFACE_TABLE::WriteFileHeaderText(ostream & out)
// write isosurface table file header
{
  out << FileHeader() << endl;
}

void ISOSURFACE_TABLE::WriteTableText(ostream & out)
// write isosurface table in text format
{
  for (int it = 0; it < NumTableEntries(); it++) {
    out << NumSimplices(it) << endl;
    for (int is = 0; is < NumSimplices(it); is++) {
      for (int iv = 0; iv < NumVerticesPerSimplex(); iv++) {
	out << "  " << SimplexVertex(it, is, iv);
      };
      out << endl;
    };
  };
}

void ISOSURFACE_TABLE::ReadText(istream & in)
// read polyhedron & table, text format
{
  int d;
  int num_poly_vert;
  int num_poly_edges;

  if (!in)
    throw ISOSURFACE_TABLE_ERROR("Input stream not open.");

  ReadFileHeaderText(in);

  in >> d;
  if (d < 2)
    throw ISOSURFACE_TABLE_ERROR("Dimension must be at least 2.");

  polyhedron.SetDimension(d);

  in >> num_poly_vert;
  if (num_poly_vert < 1)
    throw ISOSURFACE_TABLE_ERROR
      ("Number of polyhedron vertices must be at least 1.");
    
  polyhedron.SetNumVertices(num_poly_vert);
  polyhedron.ReadVertexCoordText(in);

  in >> num_poly_edges;
  if (num_poly_edges < 1)
    throw ISOSURFACE_TABLE_ERROR
      ("Number of polyhedron edges must be at least 1.");

  polyhedron.SetNumEdges(num_poly_edges);
  polyhedron.ReadEdgeListText(in);

  AllocTable();

  ReadTableText(in);
}

void ISOSURFACE_TABLE::ReadFileHeaderText(istream & in)
// read isosurface table file header in text format
{
  char c;

  if (!in)
    throw ISOSURFACE_TABLE_ERROR
      ("Error reading isosurface table.");

  while (in.get(c))
    {
      if (!isspace(c)) {
	break;
      };
    };

  if (in.eof())
    throw ISOSURFACE_TABLE_ERROR("Empty file.");
  if (!in)
    throw ISOSURFACE_TABLE_ERROR
      ("Error reading isosurface table.");

  int buffer_length = strlen(FileHeader())+1;
  char * buffer = new char[buffer_length];

  int i = 0;
  while ((i < buffer_length-1) && !isspace(c) && in) {
    buffer[i] = c;
    i++;
    if (!in.get(c))
      break;
  };
  buffer[i] = '\0';

  if (strcmp(FileHeader(), buffer) != 0) {
    delete [] buffer;
    throw ISOSURFACE_TABLE_ERROR("Illegal file header.");
  };

  delete [] buffer;
}

void ISOSURFACE_TABLE::ReadTableText(istream & in)
// read isosurface table in text format
{
  for (int it = 0; it < NumTableEntries(); it++) {

    if (in.eof())
      throw ISOSURFACE_TABLE_ERROR("Premature end of file.");
    if (!in)
      throw ISOSURFACE_TABLE_ERROR
	("Error reading isosurface table.");

    int nums;
    in >> nums;
    if (nums < 0)
      throw ISOSURFACE_TABLE_ERROR
	("Number of simplices cannot be negative.");

    SetNumSimplices(it, nums);
    for (int is = 0; is < NumSimplices(it); is++) {
      for (int iv = 0; iv < NumVerticesPerSimplex(); iv++) {
	int ie;
	in >> ie;
	if (ie < 0 || ie >= polyhedron.NumEdges())
	  throw ISOSURFACE_TABLE_ERROR
	    ("Illegal edge index in isosurface table.");

	SetSimplexVertex(it, is, iv, ie);
      };
    };
  };
}

bool ISOSURFACE_TABLE::CheckDimension(const int d) const
// check dimension
{
  if (d < 1)
    return(false);
  else
    return(true);
}

bool ISOSURFACE_TABLE::CheckTable() const
// check table
{
  if (polyhedron.NumVertices() >= LONG_BIT || polyhedron.NumVertices() < 1)
    return(false);

  if (polyhedron.NumVertices() > MaxNumVertices())
    return(false);

  if (NumTableEntries() != (1L << polyhedron.NumVertices()))
    return(false);

  if (entry == NULL)
    return(false);

  for (int it = 0; it < NumTableEntries(); it++)
    if (!entry[it].Check())
      return(false);

  for (int jt = 0; jt < NumTableEntries(); jt++)
    for (int is = 0; is < NumSimplices(jt); is++)
      for (int iv = 0; iv < NumVerticesPerSimplex(); iv++) {
	int ie = SimplexVertex(jt, is, iv);
	if (ie < 0 || ie >= polyhedron.NumEdges())
	  return(false);
      };

  return(true);
}
